!+
! Module: omregion_mod
!
! Authors:  Tim Harries - MSSL
!           Jon Rainger - MSSL
!           Vladimir Yershov - MSSL 
!
! Description: Calculates evselect region files for an OM source
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History:
!
!      14.05.2001 VNY - variables with the OSW origin have been added (window0, windowX0, windowY0, windowX1, windowY1)
!      14.05.2001 VNY - necessary attributes have been added to the source and background region tables
!      28.05.2001 VNY - windowX0 and windowY0 have been added to the coordinates of the avoid circle
!      02.08.2001 VNY - just testing the application of a fixed radius to the source 
!      06.08.2001 VNY - attempting to use a fixed radius of the source region 
!      07.08.2001 VNY - meaning of srcradius parameter has been changed: now it is in pixels if it is negative.
!                       In this case the radius is taken fixed as ABS(SRCRADIUS). If the paraneter 
!                       is positive, its meaning is as before - the radius of the source region 
!                       expressed in the units of FWHMs. 
!      19.09.2001 VNY - change src radius for the FAST mode
!      19.10.2001 VNY - dependencies updated

module omregion_mod

  use sasdal
  use dal
  use param  
  use ascregion_mod

  implicit none 

  private 

  public omregion

contains 

  !---------------------------------------------------------------------
  subroutine omregion
  !---------------------------------------------------------------------

    character(len=255)   :: set                                     ! Input source list.
    character(len=255)   :: srcRegion                               ! Output FITS source region file.
    character(len=255)   :: bkgRegion                               ! Output FITS background region file.
    real(kind=single) :: nFWHM                                      ! The limiting source extent in units of fwhm.
    real(kind=single) :: bkgInner                                   ! Inner radius of the background annulus in units of source region radius.
    real(kind=single) :: bkgOuter                                   ! Outer radius of the background annulus in units of source region radius.
    real(kind=single) :: srcRadius                                  ! Radius of source extraction region in units of FWHM of the PSF.
    integer(kind=int32) :: srcNumber                                ! Number of source in source list.

    type(DataSetT) :: regDataSetH                                   ! Handle on region data set
    type(DataSetT) :: listDataSetH                                  ! Handle on input list data set                     
    type(TableT) :: tableSetH                                      ! Handle on table
    type(ColumnT) :: xColH, yColH, FWHMColH                         ! Handles on columns.


!--- Added by V.Ye. 14.05.2001 - origin of the OSW
    character(len=255) :: window0                                   ! Origin of the OSW
    integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1   ! Coordinates of the OSW origin
    character(len=1) :: ch1, ch2, ch3, ch4, ch5
!---

    real(kind=single) :: r1
    real(kind=single) :: rInner
    real(kind=single) :: rOuter

    integer(kind=int32) :: maxSources
    real(kind=Single), pointer :: xPos(:)
    real(kind=Single), pointer :: yPos(:)
    real(kind=Single), pointer :: FWHMmajor(:)

    real(kind=Single) :: d
    real(kind=Single) :: minDist
    real(kind=Single) :: dist
    integer(kind=int32) :: iMin
    integer(kind=int32) :: i

    integer(kind=int32) , parameter :: maxAvoid = 100
    integer(kind=int32) :: nCont
    real(kind=Single) :: xAvoid(maxAvoid)
    real(kind=Single) :: yAvoid(maxAvoid)
    real(kind=Single) :: rAvoid(maxAvoid)


! Get parameters.

    set = stringParameter("set")
    srcNumber = intParameter("srcnumber")
    nFWHM = realParameter("nfwhm")
    bkgInner = realParameter("bkginner")
    bkgOuter = realParameter("bkgouter")
    srcRadius = realParameter("srcradius")
    srcRegion = stringParameter("srcfile")
    bkgRegion = stringParameter("bkgfile")


! Get handle on source list.

    listDataSetH = dataset(set, read)
    tableSetH = table(listDataSetH, 0)
    maxSources = numberOfRows(tableSetH)

! Check source number is valid.
    if (srcNumber > maxSources) call error("tooBig","Source number too big")


    srcNumber = srcNumber - 1  ! dal indices start at zero

    xColH = column(tableSetH, "XPOS", read)
    xPos => real32Data(xColH)

    yColH = column(tableSetH, "YPOS", read)
    yPos => real32Data(yColH)


! Since the regions are currently just circular, we need to use the major axis of the ellipse.

    FWHMColH = column(tableSetH, "FWHM_MAJ", read)
    FWHMmajor => real32Data(FWHMColH)

! Set the radius.


!*******************************************************************************************************************
!   VNY 07.08.2001 If the source radius parameter is negative, then apply it directly (in pixels). If it is positive, 
!                  the meaning of the parameter is number of FWHMs around the source.
!*******************************************************************************************************************
    if (srcRadius .gt. 0.0) then
       r1 = srcRadius * FWHMmajor(srcNumber)
    else
       r1= abs(srcRadius)
    end if
!---

! Find the nearest neighbour.

    minDist = HUGE(minDist)

    iMin = 0
    do i = 0 , maxSources-1
       if (i /= srcNumber) then
          dist = (xPos(srcNumber)-xPos(i))**2 + (yPos(srcNumber)-yPos(i))**2
          dist = sqrt(dist)
          if (dist < minDist) then 
             minDist = dist
             iMin = i
          endif
       endif
    enddo


! If the nearest neighbour is too close to the source then we must reduce the size of the source radius.

    if (r1 > (minDist-nFWHM*FWHMmajor(iMin))) then
!---VNY 19.09.2001 For the fixed radius (srcRadius<0) the reduction is not applied (FAST-mode)
       if (srcRadius .gt. 0.0) then
          r1 = minDist  - nFWHM * FWHMmajor(iMin)
          call warning("reducing","reducing radius of source region")
          if (r1 < 0) then
             call warning("contaminated", &
                  "Source region is unavoidably contaminated")
             r1 = FWHMmajor(srcNumber)
          endif
       else
          call warning("contaminated", "there is another source nearby")
       end if
    endif


! Set the background annulus radii.

    rInner = bkgInner * r1
    rOuter = bkgOuter * r1
    nCont = 0


! Find sources that are within the annulus and add them to a list of sources to be avoided.

    do i = 0 , maxSources-1
       if ( i /= srcNumber) then
          dist = (xPos(srcNumber)-xPos(i))**2 + (yPos(srcNumber)-yPos(i))**2
          dist = sqrt(dist)
          d = dist - nFWHM * FWHMmajor(i)

          if (d < rOuter) then
             nCont = nCont + 1
             xAvoid(nCont) = xPos(i)
             yAvoid(nCont) = yPos(i)
             rAvoid(nCont) = nFWHM * FWHMmajor(i)
          endif
       endif
    enddo


!--- Check by V.Ye. 14.05.2001 (origin of the OSW)
window0 = stringAttribute(listDataSetH, "DETSEC")

!--- 15.05.2001 V.Ye.: the origin of the OSW has been added in order to be able to extract time series
read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)') ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5


! Write the source region file.

    regDataSetH = createRegion(srcRegion)

    call copyAllAttributes(attributable(listDataSetH), attributable(regDataSetH))
    call setAttribute(regDataSetH, "SRCNO", srcNumber+1)

    call addCircle(regDataSetH, xPos(srcNumber)+windowX0, yPos(srcNumber)+windowY0, r1, 1, .true.)

! --- Added by V.Ye. 14.05.2001 (Adding necessary attributes to the source region table)
    tableSetH=table(regDataSetH,0)
    call setAttribute(tableSetH, "EXTVER", 1)
    call setAttribute(tableSetH, "EXTLEVEL", 1, "Level in DB hierarchy: Data table")
    call setAttribute(tableSetH, "CHECKSUM", "0000000000000000", "ASCII encoded HDU checksum")
    call setAttribute(tableSetH, "DATASUM", "       0", "Data unit checksum written in ASCII")
    call setAttribute(tableSetH, "ORIGIN", "OMREGION")
    call setAttribute(tableSetH, "CONTENT", "REGION ")
    call setAttribute(tableSetH, "HDUDOC", "Unknown")
    call setAttribute(tableSetH, "HDUVERS", "1.0.0")
    call setAttribute(tableSetH, "HDUCLASS", "ASC     ")
    call setAttribute(tableSetH, "HDUCLAS2", "STANDARD")
    call setAttribute(tableSetH, "MTYPE1", "pos     ")
! ---


    call closeRegion(regDataSetH)

! Write the background region file.
    
    regDataSetH = createRegion(bkgRegion)

    call copyAllAttributes(attributable(listDataSetH), attributable(regDataSetH))
    call addAnnulus(regDataSetH, xPos(srcNumber)+windowX0, yPos(srcNumber)+windowY0, &
                    rInner, rOuter, 1, .true.)

! All these circles will be excluded.

    do i = 1 , nCont
       call addCircle(regDataSetH, xAvoid(i)+windowX0, yAvoid(i)+windowY0, rAvoid(i), 1, .false.)
    enddo

! --- Added by V.Ye. 14.05.2001 (Adding necessary attributes to the background  region table)
    tableSetH=table(regDataSetH,0)
    call setAttribute(tableSetH, "EXTVER", 1)
    call setAttribute(tableSetH, "EXTLEVEL", 1, "Level in DB hierarchy: Data table")
    call setAttribute(tableSetH, "CHECKSUM", "0000000000000000", "ASCII encoded HDU checksum")
    call setAttribute(tableSetH, "DATASUM", "       0", "Data unit checksum written in ASCII")
    call setAttribute(tableSetH, "ORIGIN", "OMREGION")
    call setAttribute(tableSetH, "CONTENT", "REGION ")
    call setAttribute(tableSetH, "HDUDOC", "Unknown")
    call setAttribute(tableSetH, "HDUVERS", "1.0.0")
    call setAttribute(tableSetH, "HDUCLASS", "ASC     ")
    call setAttribute(tableSetH, "HDUCLAS2", "STANDARD")
    call setAttribute(tableSetH, "MTYPE1", "pos     ")
! ---

! Release the dal handles.

    call closeRegion(regDataSetH)


    call release(xColH)
    call release(yColH)
    call release(FWHMColH)

    call release(listDataSetH)
  end subroutine omregion

end module omregion_mod



